#!/usr/bin/env python3

#
# Swift Navigation PGM Programming Tool
#
# MIT License
# Copyright (c) 2020 by Swift Navigation, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# std lib
from abc import ABCMeta
import argparse
import binascii
import json
import logging
import os
import os.path
from pathlib import Path
import subprocess
import struct
import sys
import time
import traceback
from typing import Optional
from zipfile import ZipFile

try:
    from fcntl import ioctl
except ModuleNotFoundError:
    pass  # Windows

try:
    import serial
except ModuleNotFoundError as e:
    if __name__ == "__main__":
        raise e
    else:
        print(
            "Warning, PySerial Not Found when importing pgm_update.\n"
            "Supressing the following exception:\n" + traceback.format_exc()
        )
        pass

try:
    import stm32loader.main

    STM32LOADER_IMPORTED = True
except ModuleNotFoundError as e:
    STM32LOADER_IMPORTED = False

LOGGER = logging.getLogger(__name__)

# linux defaults
DEFAULT_AUX_DEV = "/dev/pgm1aux"
DEFAULT_RESET_DEV = "/dev/pgm1reset"
DEFAULT_BOOT_DEV = "/dev/pgm1boot"

PPT_VERSION = "3.0.2"


SYNC_PATTERNS = (
    (b"\x09\x13\x4c\x51", b"\x4d\x3a\xfc\xaa"),
    (b"\x04\xa5\x03\x12", b"\xa0\x5b\xfd\x55"),
)

FRAME_PREFIX = b"\xaa"
FRAME_SUFFIX = b"\x55"

CHUNKSIZE = 0x1400

TARGET_STM32_BOOTLOADER = 0
TARGET_STA8100 = 1
TARGET_STM32 = 2

CMD_ACK = 0
CMD_ADDRESS = 1
CMD_BIN_INFO = 2
CMD_START = 3
CMD_DOWNLOAD = 4

ACK_SUCCESS = 0

BIN_INFO_MAGIC = {  # LG69TFlashTool_V1.5 magic
    TARGET_STA8100: b"\x10\x00\x00\x00"  # Dest Address
    + b"\x00\x00\x04\x00"  # Entry Point
    + b"\x00\x18\x00\x00"  # NVM offset
    + b"\x00\x08\x00\x00"  # NVM size
    + b"\x01"  # Erase NVM
    + b"\x00"  # Erase only
    + b"\x00"  # Program Only
    + b"\x00\x00\x00\x00\x00",  # Reserved
    TARGET_STM32: b"\x00\x00\x00\x00\x01\x00\x00\x00",
}

# Linux-specific IOCTL stuff
GPIO_GET_LINEHANDLE_IOCTL = 0xC16CB403
GPIOHANDLE_SET_LINE_VALUES_IOCTL = 0xC040B409
GPIOHANDLE_REQUEST_OUTPUT = 1 << 1

# CP2105 pin enumeration does not appear to be consistent across
# platforms!
# All pin IDs need to be empirically confirmed.

# Pins associated with Standard COM port / bank
RESET_PIN_ID = {"linux": 1, "macos": 3, "win": 1}
BATT_OFF_PIN_ID = {"linux": 0, "macos": 0, "win": 0}  # TODO: mac values
LG69TPWR_OFF_PIN_ID = {"linux": 2, "macos": 0, "win": 2}  # TODO: macos values

# Pins associated with Enhanced COM port / bank
BOOT_PIN_ID = {"linux": 0, "macos": 0, "win": 0}

PGM_RESET_DEFAULT_DIR_MACOS = Path("tools/pgm_reset/build_Darwin")
PGM_RESET_DEFAULT_DIR_WINDOWS = Path("tools/pgm_reset/build_Windows")

PGM_RESET_DURATION = 1
POST_QT_SYNC_DELAY = 1
PGM_POWER_OFF_DURATION = 2
PGM_POWER_ON_DELAY = 2

# CLI stuff
CLI_TARGETS = {
    "stm32_bootloader": TARGET_STM32_BOOTLOADER,
    "stm32": TARGET_STM32,
    "sta8100": TARGET_STA8100,
    "combined": None,
}

TARGET_LABELS = {0: "stm32_bootloader", 1: "sta8100", 2: "stm32"}


class PgmResetToolError(Exception):
    """
    Exception specific to pgm_reset(.exe) tool so we can suppress the
    FF-433 false failure.
    """


class PgmResetMixin:
    def _invoke_pgm_reset(self, command: str) -> None:
        try:
            subprocess.run(command, check=True, shell=True, capture_output=True)
        except subprocess.CalledProcessError as err:
            # handling that is specific to pgm_reset tool
            msg = err.output.decode("ascii").strip().split("\n")[-1]
            raise PgmResetToolError(msg) from err


class CP2105OutputPin(metaclass=ABCMeta):
    def set(self, value: int):
        raise NotImplementedError("concrete class must implement set() method")


class CP2105OutputPinLinux(CP2105OutputPin):
    """SiLabs CP2105 GPIO output pin control abstraction on Linux.
    """

    def __init__(
        self,
        dev_path: str,
        pin: int,
        initial_value: int = 0,
        label: str = "python-gpio",
    ):
        if not sys.platform.startswith("linux"):
            raise RuntimeError(
                "generic CP2105 GPIO pin control is only available on Linux"
            )

        # send a request to the /dev mountpoint and
        # receive a file descriptor for the requested pin
        fd = os.open(dev_path, 0)
        req = struct.pack("=64L", pin, *bytearray(63))
        req += struct.pack("=L", GPIOHANDLE_REQUEST_OUTPUT)
        req += struct.pack("=64B", initial_value, *bytearray(63))
        req += struct.pack("=32s", bytes(label, "utf-8"))
        req += struct.pack("=LL", 1, 0)
        ret = ioctl(fd, GPIO_GET_LINEHANDLE_IOCTL, req)
        self.fd = struct.unpack("=L", ret[-4:])[0]
        os.close(fd)

        self.pin = pin
        self.dev = dev_path

    def set(self, value: int):
        msg = "instructing pin {pin} ({dev}) to output {num} ({human})"
        msg = msg.format(
            pin=self.pin, dev=self.dev, num=value, human="high" if value else "low"
        )
        LOGGER.info(msg)

        data = struct.pack("=64B", value, *bytearray(63))
        ioctl(self.fd, GPIOHANDLE_SET_LINE_VALUES_IOCTL, data)


class CP2105OutputPinMacOS(PgmResetMixin, CP2105OutputPin):
    """SiLabs CP2105 GPIO output pin control abstraction on MacOS.
    Leverages *pgm_reset* utility.
    """

    def __init__(self, usb_loc_id: int, pin: int, tool_dir: Path):
        self.usb_loc_id = usb_loc_id
        self.pin = pin
        self.tool_dir = tool_dir

    def set(self, value) -> None:
        if value == 0:
            mac_value = 1
        else:
            mac_value = 0

        msg = "instructing pin {pin} (USB LocID {dev}) to output {num} ({human})"
        msg = msg.format(
            pin=self.pin,
            dev=self.usb_loc_id,
            num=mac_value,
            human="high" if value else "low",
        )
        LOGGER.info(msg)

        invocation = "LD_LIBRARY_PATH={lib} {exe} {usb} manual {pin}={value}"
        command = invocation.format(
            lib=self.tool_dir,
            exe=self.tool_dir / "pgm_reset",
            usb=self.usb_loc_id,
            pin=self.pin,
            value=mac_value,
        )
        self._invoke_pgm_reset(command)


class CP2105OutputPinWindows(PgmResetMixin, CP2105OutputPin):
    """SiLabs CP2105 GPIO output pin control abstraction on Windows.
    Leverages *pgm_reset* utility.
    """

    def __init__(self, com_port: str, pin: int, tool_dir: Path):
        self.com_port = com_port
        self.pin = pin
        self.tool_dir = tool_dir

    def set(self, value) -> None:
        msg = "instructing pin {pin} ({dev}) to output {num} ({human})"
        msg = msg.format(
            pin=self.pin, dev=self.com_port, num=value, human="high" if value else "low"
        )
        LOGGER.info(msg)

        # double-quoting path to exe to handle modern Windows paths that
        # contain spaces
        invocation = '"{exe}" {com} manual {pin}={value}'
        command = invocation.format(
            exe=self.tool_dir / "pgm_reset",
            com=self.com_port,
            pin=self.pin,
            value=value,
        )
        self._invoke_pgm_reset(command)


class CP2105OutputPinFactory:
    def __init__(
        self,
        pin_id_linux: int,
        pin_id_macos: int,
        pin_id_win: int,
        tool_dir: Optional[Path] = None,
        initial_value: int = 0,
        label: str = "python-gpio",
    ):
        self.pin_id_linux = pin_id_linux
        self.pin_id_macos = pin_id_macos
        self.pin_id_win = pin_id_win
        self.initial_value = initial_value
        self.label = label
        self.tool_dir = tool_dir

    @staticmethod
    def _get_linux_pin(dev_com: str, pin_id: int, initial_value: int, label: str):
        return CP2105OutputPinLinux(dev_com, pin_id, initial_value, label)

    @staticmethod
    def _get_macos_pin(usb_loc_id: int, pin_id: int, tool_dir: Optional[Path]):
        return CP2105OutputPinMacOS(usb_loc_id, pin_id, tool_dir)

    @staticmethod
    def _get_win_pin(dev_com: str, pin_id: int, tool_dir: Optional[Path]):
        return CP2105OutputPinWindows(dev_com, pin_id, tool_dir)

    def get_pin(
        self, lnx_win_dev: Optional[str], mac_usb_loc: Optional[int],
    ) -> CP2105OutputPin:
        if sys.platform.startswith("linux"):
            return self._get_linux_pin(
                lnx_win_dev, self.pin_id_linux, self.initial_value, self.label
            )
        elif sys.platform.startswith("darwin"):
            if not self.tool_dir:
                self.tool_dir = PGM_RESET_DEFAULT_DIR_MACOS
            return self._get_macos_pin(mac_usb_loc, self.pin_id_macos, self.tool_dir)
        elif sys.platform.startswith("win"):
            if not self.tool_dir:
                self.tool_dir = PGM_RESET_DEFAULT_DIR_WINDOWS
            return self._get_win_pin(lnx_win_dev, self.pin_id_win, self.tool_dir)
        else:
            msg = "{} is not a supported operating system"
            msg = msg.format(sys.platform)
            raise RuntimeError(msg)


class LG69TArchive(ZipFile):
    """Abstraction of a zip archive that packages the TeseoV and STM32
    firmware binaries required by the Quectel LG69T into a single file.
    """

    def __init__(self, file, mode="r"):
        ZipFile.__init__(self, file, mode)
        self.mode = mode
        self.closed = False
        if mode == "r":
            self.manifest = json.loads(self.read("manifest.json"))
        else:
            self.manifest = []

    def write(self, target, file, magic=None):
        arcname = os.path.basename(file)
        manifest = {"target": target, "binary": arcname}
        if magic:
            manifest["magic"] = magic.hex()
        ZipFile.write(self, file, arcname)
        self.manifest.append(manifest)

    def __iter__(self):
        for x in self.manifest:
            target = x["target"]
            data = self.read(x["binary"])
            magic = bytes.fromhex(x["magic"]) if "magic" in x else None
            yield target, data, magic

    def close(self):
        if getattr(self, "closed", False):
            return
        self.closed = True
        if self.mode == "w":
            manifest = json.dumps(self.manifest, indent="\t")
            self.writestr("manifest.json", manifest)
        ZipFile.close(self)


class LG69TBootloader(object):
    """Abstraction of bootloader protocol observed thus far in Quectel
    LG69T.
    """

    def __init__(self, port):
        self._port = port
        self._bininfo = BIN_INFO_MAGIC

    def set_bininfo(self, target, magic):
        self._bininfo[target] = magic

    def send_frame_wait(self, target, cmd, data, timeout=5, retries=2):
        while retries >= 0:
            self.send_frame(target, cmd, data)
            if self.wait_ack(target, cmd, timeout):
                return
            retries -= 1
            print("Timeout waiting for ack")
        raise Exception("No ack received")

    def send_frame(self, target, cmd, data):
        header = struct.pack(">BBH", target, cmd, len(data))
        payload = header + data
        crc = struct.pack(">L", binascii.crc32(payload) & 0xFFFFFFFF)
        self._port.write(FRAME_PREFIX + payload + crc + FRAME_SUFFIX)

    def wait_ack(self, target, cmd, timeout):
        expiry = time.time() + timeout
        waits = 0
        while time.time() < expiry:
            if waits != 0 and waits % 6 == 0:
                sys.stdout.write(".")
                sys.stdout.flush()
            ack = self.recv_frame()
            if ack is not None:
                if waits >= 6:
                    print()
                break
            waits += 1
        if ack is None:
            return False

        at, ac, payload = ack
        if (at != target) or (ac != CMD_ACK):
            raise Exception(
                "Invalid ack: at {} ac {} payload {}".format(at, ac, payload)
            )

        rt, rc, code = struct.unpack(">BBH", payload)
        if (rt != target) or (rc != cmd) or (code != ACK_SUCCESS):
            raise Exception("Invalid ack (%x %x %x)" % (rt, rc, code))
        return True

    def recv_frame(self):
        prefix = self._port.read(1)
        if prefix != FRAME_PREFIX:
            return None
        header = self._port.read(4)
        target, cmd, msgsize = struct.unpack(">BBH", header)
        payload = self._port.read(msgsize)
        crc = self._port.read(4)
        crc = struct.unpack(">L", crc)[0]
        suffix = self._port.read(1)
        if crc != binascii.crc32(header + payload) & 0xFFFFFFFF:
            print("CRC error %x != %x" % (crc, binascii.crc32(header + payload)))
            return None
        if suffix != FRAME_SUFFIX:
            print("ignoring incorrect suffix")
        return target, cmd, payload

    def sync(self):
        print("Waiting for sync...")
        for request, response in SYNC_PATTERNS:
            while True:
                self._port.write(request)
                data = self._port.read(1024)
                if data[-4:] == response:
                    break
        print("Synchronisation success")

    def send_address(self, target, address=0):
        self.send_frame_wait(target, CMD_ADDRESS, struct.pack(">L", address))
        print("Send address success")

    def binimage_crc(self, binary):
        # CRC-32 over the length encoded as LE32 and then over the full binary
        crc = binascii.crc32(struct.pack("<L", len(binary)))
        return binascii.crc32(binary, crc) & 0xFFFFFFFF

    def send_binary_info(self, target, binary):
        size = len(binary)
        crc = self.binimage_crc(binary)

        bininfo = struct.pack(">LL", size, crc) + self._bininfo[target]

        self.send_frame_wait(target, CMD_BIN_INFO, bininfo)
        print("Send binary information success")

    def start(self, target):
        self.send_frame_wait(target, CMD_START, b"", timeout=180)
        print("Send start command success")

    def download(self, target, binary):
        chunks = (len(binary) + CHUNKSIZE - 1) // CHUNKSIZE
        for chunk in range(chunks):
            data = binary[chunk * CHUNKSIZE : (chunk + 1) * CHUNKSIZE]
            print(
                "Programming %s, block %d (%.1f%%)"
                % (target, chunk, 100.0 * chunk / chunks),
                flush=True,
            )
            if chunk == (chunks - 1):  # last packet long timeout
                self.send_frame_wait(
                    target, CMD_DOWNLOAD, struct.pack(">L", chunk) + data, timeout=180
                )
            else:
                self.send_frame_wait(
                    target, CMD_DOWNLOAD, struct.pack(">L", chunk) + data
                )

        print("Programming complete")

    def prog(self, target, binary):
        self.send_address(target)
        self.send_binary_info(target, binary)
        self.start(target)
        self.download(target, binary)


class UpdaterPromptParser(argparse.ArgumentParser):
    def __init__(self):
        super().__init__(description="Swift Navigation PGM Programming Tool")
        self.add_argument(
            "-b",
            "--baud-rate",
            default=460800,
            help="baud rate, default: 460800",
            dest="baud",
            type=int,
        )
        self.add_argument(
            "-t",
            "--target",
            default="combined",
            choices=CLI_TARGETS.keys(),
            help="target to program, default: combined",
            dest="target",
        )
        self.add_argument(
            "-s", "--no-sync", action="store_true", help="skip sync", dest="no_sync"
        )
        self.add_argument("binary", nargs=1, help="binary file or combined archive")
        self.add_argument(
            "--version", action="version", version="",
        )

        if sys.platform.startswith("linux"):
            self.add_argument(
                "-p",
                "--serial-port",
                default=DEFAULT_AUX_DEV,
                help="AUX serial port, default: {}".format(DEFAULT_AUX_DEV),
                dest="data_dev",
            )
            self.add_argument(
                "--boot-pin",
                default=DEFAULT_BOOT_DEV,
                help="boot pin GPIO device, default: {}".format(DEFAULT_BOOT_DEV),
                dest="boot_dev",
            )
            self.add_argument(
                "-g",
                "--gpio-reset",
                default=DEFAULT_RESET_DEV,
                help="GPIO device for reset, default: {}".format(DEFAULT_RESET_DEV),
                dest="reset_dev",
            )
        elif sys.platform.startswith("darwin"):
            self.add_argument(
                metavar="usb_location_id",
                help="PGM USB location ID",
                dest="usb_loc_id",
            )
            self.add_argument(
                metavar="cu_slab_enhanced", help="serial port", dest="data_dev"
            )
        elif sys.platform.startswith("win"):
            self.add_argument(
                metavar="standard_com", help="Standard COM port", dest="reset_dev"
            )
            self.add_argument(
                metavar="enhanced_com", help="Enhanced COM port", dest="data_dev"
            )
        else:
            msg = "{} is not supported".format(sys.platform)
            raise RuntimeError(msg)


def reset_pgm(gpio: CP2105OutputPin):
    try:
        gpio.set(1)
        time.sleep(PGM_RESET_DURATION)
        gpio.set(0)
    except subprocess.CalledProcessError as err:
        if err.output:
            msg = err.output.decode("ascii", "replace").strip().split("\n")
        else:
            msg = ""

        msg = msg[-1] if msg else ""
        raise RuntimeError(msg) from err


def validate_binary_path(parsed_arg: str) -> Path:
    binary_path = Path(parsed_arg)

    if not binary_path.is_file():
        # also tests for existence
        # also works on symlinks according to docs
        raise FileNotFoundError(
            "`{}` is not a valid firmware archive or binary file".format(binary_path)
        )

    return binary_path


def cts_is_asserted(uart_dev: str, baud_rate: str) -> bool:
    """Checks if CTS is asserted (pulled low).

    Returns
    -------
    bool:
        ``True`` for 01097-03 Rev01 and later. ``False`` for earlier
        hardware revisions.
    """
    with serial.Serial(uart_dev, baud_rate, timeout=0.1) as uart_port:
        is_asserted = uart_port.cts
    return is_asserted


def disconnect_battery_and_powercycle(
    batt_pin: CP2105OutputPin, power_pin: CP2105OutputPin
) -> None:
    # battery and power pins are active low
    batt_pin.set(0)

    power_pin.set(0)
    time.sleep(PGM_POWER_OFF_DURATION)
    power_pin.set(1)
    time.sleep(PGM_POWER_ON_DELAY)


def reconnect_battery(batt_pin: CP2105OutputPin) -> None:
    try:
        # battery pin is active low
        batt_pin.set(1)
    except PgmResetToolError as err:
        LOGGER.debug("ignoring %s (FF-433 false failure)", err)


def access_stm32_rom_bootloader(
    reset_pin: CP2105OutputPin, boot_pin: CP2105OutputPin
) -> None:
    boot_pin.set(0)
    reset_pgm(reset_pin)
    time.sleep(1)
    boot_pin.set(1)


def update_stm32_bootloader(
    reset_pin: CP2105OutputPin,
    boot_pin: CP2105OutputPin,
    data_dev: str,
    baud_rate: str,
    bin_path: Path,
) -> None:
    if not STM32LOADER_IMPORTED:
        msg = "Updating target (-t) `stm32_bootloader` requires 3rd-party module `stm32loader`. Aborting."
        raise RuntimeError(msg)

    STM32_LOAD_CONFIG = {
        "port": None,
        "baud": 115200,
        "parity": stm32loader.main.Stm32Loader.PARITY["even"],
        "family": "H7",
        "address": 0x08000000,
        "erase": True,
        "unprotect": False,
        "write": True,
        "verify": True,
        "read": False,
        "go_address": -1,
        "swap_rts_dtr": False,
        "reset_active_high": False,
        "boot0_active_low": False,
        "hide_progress_bar": False,
        "data_file": None,
    }
    access_stm32_rom_bootloader(reset_pin, boot_pin)

    loader = stm32loader.main.Stm32Loader()
    loader.configuration = STM32_LOAD_CONFIG
    loader.configuration["data_file"] = bin_path
    loader.configuration["port"] = data_dev
    loader.configuration["baud"] = int(baud_rate)
    loader.connect()
    loader.read_device_id()
    loader.read_device_uid()
    loader.perform_commands()

    print("Programming complete")
    print("Successfully programmed stm32_bootloader")


def access_lg69t_bootloader(
    data_dev: str, baud_rate: str, reset_pin: CP2105OutputPin, no_sync: bool = False
) -> LG69TBootloader:
    enhanced_serial_port = serial.Serial(data_dev, baud_rate, timeout=0.1)
    bootloader = LG69TBootloader(enhanced_serial_port)

    if not no_sync:
        reset_pgm(reset_pin)
        bootloader.sync()
        time.sleep(POST_QT_SYNC_DELAY)

    return bootloader


def update_pgm_firmware_binary(
    data_dev: str,
    baud_rate: str,
    reset_pin: CP2105OutputPin,
    bin_path: Path,
    target: str,
    no_sync: bool = False,
) -> None:
    bootloader = access_lg69t_bootloader(data_dev, baud_rate, reset_pin, no_sync)

    binary = open(bin_path, "rb").read()
    bootloader.prog(CLI_TARGETS[target], binary)

    print("Successfully programmed {}".format(target))


def update_pgm_firmware_combined(
    data_dev: str,
    baud_rate: str,
    reset_pin: CP2105OutputPin,
    bin_path: Path,
    no_sync: bool = False,
) -> None:
    bootloader = access_lg69t_bootloader(data_dev, baud_rate, reset_pin, no_sync)

    for target, binary, magic in LG69TArchive(str(bin_path), "r"):
        if magic:
            bootloader.set_bininfo(target, magic)
        bootloader.prog(target, binary)

        target_label = TARGET_LABELS[target]
        print("Successfully programmed {}".format(target_label))


def pretty_print_exception(exc: Exception) -> None:
    print("\n--- << DEBUG >> ----------------------------------")
    traceback.print_exc()
    print("---------------------------------- << DEBUG >> ---\n")
    print("Programming failed")
    print(exc)


def update_pgm_firmware(tool_dir: Optional[Path] = None):
    try:
        print("Swift Navigation PGM Programming Tool v{}".format(PPT_VERSION))

        parser = UpdaterPromptParser()
        args = parser.parse_args()
        if "reset_dev" not in args:
            args.reset_dev = None
        if "usb_loc_id" not in args:
            args.usb_loc_id = None

        reset_pin_factory = CP2105OutputPinFactory(
            RESET_PIN_ID["linux"], RESET_PIN_ID["macos"], RESET_PIN_ID["win"], tool_dir
        )
        boot_pin_factory = CP2105OutputPinFactory(
            BOOT_PIN_ID["linux"], BOOT_PIN_ID["macos"], BOOT_PIN_ID["win"], tool_dir
        )
        batt_pin_factory = CP2105OutputPinFactory(
            BATT_OFF_PIN_ID["linux"],
            BATT_OFF_PIN_ID["macos"],
            BATT_OFF_PIN_ID["win"],
            tool_dir,
        )
        power_pin_factory = CP2105OutputPinFactory(
            LG69TPWR_OFF_PIN_ID["linux"],
            LG69TPWR_OFF_PIN_ID["macos"],
            LG69TPWR_OFF_PIN_ID["win"],
            tool_dir,
        )

        bin_path = validate_binary_path(args.binary[0])

        reset_pin = reset_pin_factory.get_pin(
            lnx_win_dev=args.reset_dev, mac_usb_loc=args.usb_loc_id
        )

        is_01097_03_hw = cts_is_asserted(args.data_dev, args.baud)
        if is_01097_03_hw:
            print("Identified hardware as 01097-03 Rev01 or later...")
            # only 01097-03 Rev01 and later have the BATT_OFF and
            # LG69TPWR_OFF pins for disconnecting the backup battery
            batt_pin = batt_pin_factory.get_pin(
                # same bank as RESET pin
                # shares the same /dev/... path
                lnx_win_dev=args.reset_dev,
                mac_usb_loc=args.usb_loc_id,
            )
            power_pin = power_pin_factory.get_pin(
                # same bank as RESET pin
                # shares the same /dev/... path
                lnx_win_dev=args.reset_dev,
                mac_usb_loc=args.usb_loc_id,
            )
            disconnect_battery_and_powercycle(batt_pin, power_pin)
        else:
            print("Identified hardware as 01097-02 Rev04 or earlier...")

        if args.target == "stm32_bootloader":
            if sys.platform.startswith("linux"):
                boot_pin = boot_pin_factory.get_pin(
                    lnx_win_dev=args.boot_dev, mac_usb_loc=args.usb_loc_id
                )
            else:
                boot_pin = boot_pin_factory.get_pin(
                    lnx_win_dev=args.data_dev, mac_usb_loc=args.usb_loc_id
                )
            update_stm32_bootloader(
                reset_pin, boot_pin, args.data_dev, args.baud, bin_path
            )
        if args.target == "stm32":
            update_pgm_firmware_binary(
                args.data_dev, args.baud, reset_pin, bin_path, args.target, args.no_sync
            )
        if args.target == "sta8100":
            update_pgm_firmware_binary(
                args.data_dev, args.baud, reset_pin, bin_path, args.target, args.no_sync
            )
        if args.target == "combined":
            update_pgm_firmware_combined(
                args.data_dev, args.baud, reset_pin, bin_path, args.no_sync
            )

        reset_pgm(reset_pin)
        if is_01097_03_hw:
            reconnect_battery(batt_pin)
        sys.exit(0)
    except Exception as exc:
        pretty_print_exception(exc)
        sys.exit(1)


if __name__ == "__main__":
    update_pgm_firmware()